/*
 * chk4mail - mail folder check
 * $Id: chk4mail.c,v 1.16 2007/07/05 12:08:15 johans Exp $
 */

/*
 * Copyright (C) 1998 - 2008, Johan van Selst <johans@stack.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#ifndef	_GNU_SOURCE
# define	_GNU_SOURCE
#endif	/* _GNU_SOURCE */

#ifdef	__GNUC_GNU_INLINE__
# define	inline		__attribute__ ((gnu_inline)) inline
#endif	/* __GNUC_GNU_INLINE__ */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <utime.h>
#include <err.h>
#include <sys/utsname.h>
#include <paths.h>
#include <pwd.h>
#include <stdbool.h>
#include <getopt.h>

#include "strlcpy.h"

#define HEADERLEN	1001	/**< rfc2822: max 998 chars (+ crlf + null) */

#define qerr(st, ...) \
	do { if (quiet) return false; else err((st), __VA_ARGS__); } while (0)

unsigned int	spc = 5,	///< max digits for folder numbers
		mlen = 4;	///< max folder name length
char		**exclude;	///< list of filenames to ignore
bool		hideold = false, hideunread = false,
		quiet = false, brief = false, oflag = false,
		oanswer = false, printfrm = false, printfrmold = false,
		unreadasold = false, touch = false,
		showdot = false, caps = false, flink = false;

/**
 * \struct summary
 * \brief number of messages found in a mailbox (per message type)
 */
typedef struct	summary
{
	unsigned int	new;	///< number of unread mails in folder
	unsigned int	flag;	///< flagged (important) mails
	unsigned int	answer;	///< mails that you replied to
	unsigned int	total;	///< messages in total
	char		*frm;	///< text with from/subject data of the mails
	unsigned int	frmsz;	///< size of frm buffer
} summary_t;

/**
 * Print command line options and exit
 */
void
usage(void)
{
	/* give clue */
	errx(1, "chk4mail " PACKAGE_VERSION " - mail folder check\n"
"Mbox, MMDF, MH, Maildir and Extended Maildir folders are recognized.\n"
"\n"
"-# <n>       Pad the number of messages with n spaces\n"
"-a           Show how many messages have been answered\n"
"-c           Include folder names with capitals\n"
"-d <dir>     Look for mailfolders in this directory instead\n"
"-e <folder>  Exclude the specified folder\n"
"-f           Show how many messages have been flagged as important\n"
"-h           Hide folders containing only old mail\n"
"-i           Only parse standard incoming mailfolder\n"
"-j           Ignore the standard incoming mailfolder\n"
"-l           Always follow symbolic links\n"
"-m <n>       Set max folder name lenght (for layout padding)\n"
"-n           Only parse folders that received new mail\n"
"-q           Suppress errors about missing or unreadable files\n"
"-r <host>    Execute the command on an other host (with rsh)\n"
"-s           Short output shows only files with new mail, implies -n\n"
"-t           Mark old files with new messages as unread ('touch')\n"
"-u           Don't treat unread messages as new mail\n"
"-x           Include folder names with leading dots\n"
"Folders may be specified as arguments as well (filename or =folder)\n"
"Read the manual for more details.\n");
}

/**
 * Copy a string up to end-of-line
 *
 * \param[out]	dest	result buffer (at least length HEADERLEN)
 * \param[in]	src	input line, optionally containing a newline
 */
static inline void
linecopy(char *dest, const char * const src)
{
	char *t = dest;

	strlcpy(t, src, HEADERLEN);
	strsep(&t, "\r\n");
}

/**
 * Parse folder in Mbox or MMDF format
 *
 * \param[in]	file	file to parse
 * \param[out]	sum	folder summary
 * \param[in]	mmdf	is this folder in MMDF format
 *
 * \return		true if parsing succeeded
 */
bool
parse_mbox_mmdf(FILE *file, summary_t *sum, bool mmdf)
{
	unsigned int		nrs, readnum;

	nrs = readnum = 0;
	while (!feof(file))
	{
		char	line[HEADERLEN], myfrom[HEADERLEN], mysubj[HEADERLEN];
		long	skip = 0;
		bool	header = false;
		bool	isnew = true;

		/* parse headers */
		while (fgets(line, HEADERLEN, file) &&
			*line != '\r' && *line != '\n')
		{

			/* look for begin of a new message */
			if ((!mmdf && !strncmp(line, "From ", 5)) ||
				(mmdf && !strncmp(line, "\1\1\1\1", 4)))
			{
				header = true;
				myfrom[0] = mysubj[0] = '\0';
				continue;
			}
			else if (!header)
				continue;

			/* in header */
			if (!strncasecmp(line, "Content-Length: ", 16))
				skip = atol(line + 16);
			else if (!strncasecmp(line, "Status: ", 8))
			{
				if (strchr(line + 8, 'R') ||
					(unreadasold && strchr(line + 8, 'O')))
				{
					readnum++;
					isnew = 0;
				}
			}
			else if (!strncasecmp(line, "X-Status: ", 10))
			{
				if (strchr(line + 10, 'F'))
					sum->flag++;
				if (strchr(line + 10, 'A'))
					sum->answer++;
			}
			/* nfrm compatibility: guess sender name */
			else if (printfrm &&
				!strncasecmp(line, "From: ", 6))
			{
				char	dummy[HEADERLEN];
				char	*p = line + 6;

				if (sscanf(p, "\"%[^\"]\" <%s>\n", myfrom, dummy) < 2 &&
					sscanf(p, "<%s> \"%[^\"]\"\n", dummy, myfrom) < 2 &&
					sscanf(p, "%[^<] <%s>\n", myfrom, dummy) < 2 &&
					sscanf(p, "<%s> %[^<]\n", dummy, myfrom) < 2 &&
					sscanf(p, "(%[^)]) %s\n", myfrom, dummy) < 2 &&
					sscanf(p, "%s (%[^)])\n", dummy, myfrom) < 2)
				{
					linecopy(myfrom, p);
				}
			}
			/* nfrm compatibility: copy subject */
			else if (printfrm &&
					!strncasecmp(line, "Subject: ", 9))
				linecopy(mysubj, line + 9);
		}

		/* skip body (hopefully) */
		if (header)
		{
			nrs++;
			fseek(file, skip, SEEK_CUR);
			/* nfrm compatibility output */
			if (printfrmold || (printfrm && isnew))
			{
				if (strlen(sum->frm) + 80 > sum->frmsz)
				{
					sum->frmsz *= 2;
					sum->frm = realloc(sum->frm,sum->frmsz);
				}
				sprintf(sum->frm + strlen(sum->frm),
					"%-24.24s %-53.53s\n", myfrom, mysubj);
					/* len ok */
			}
		}
	}

	if (mmdf)
		nrs--;

	sum->total = nrs;
	sum->new = nrs >= readnum ? nrs - readnum : 0;
	return true;
}

/**
 * Parse folder in Maildir format (qmail or other)
 *
 * \param[in]	subdir	directory to parse
 * \param[out]	sum	folder summary
 *
 * \return		true if parsing succeeded
 */
bool
parse_maildir(const char * const subdir, summary_t *sum)
{
	char		*filename;
	struct dirent	*file;
	DIR		*dir;
	bool		isread;

	sum->new = 0;
	sum->total = 0;

	asprintf(&filename, "%s/cur", subdir);
	if (!(dir = opendir(filename)))
		qerr(1, "Cannot opendir `%s'", filename);

	while ((file = readdir(dir)))
	{
		char		*p;

		if (!strcmp(file->d_name, ".") || !strcmp(file->d_name, ".."))
			continue;
		isread = false;
		if ((p = strrchr(file->d_name, ':')) && !strncmp(p, ":2,", 3))
		{
			p += 2;
			while (*(++p))
				if ('S' == *p)
					isread = true;
				else if (oanswer && 'R' == *p)
					(sum->answer)++;
				else if (oflag && 'F' == *p)
					(sum->flag)++;
		}
		if (isread || unreadasold)
			(sum->total)++;
		else
			(sum->new)++;
	}
	closedir(dir);

	sprintf(filename, "%s/new", subdir); /* len ok */
	if (!(dir = opendir(filename)))
		qerr(1, "Cannot opendir `%s'", filename);

	while ((file = readdir(dir)))
		if (strcmp(file->d_name, ".") && strcmp(file->d_name, ".."))
			(sum->new)++;
	closedir(dir);

	sum->total += sum->new;
	free(filename);
	return true;
}

/**
 * Parse folder in MH format
 *
 * \param[in]	subdir	directory to parse
 * \param[out]	sum	folder summary
 *
 * \return		true if parsing succeeded
 */
bool
parse_mh(const char *subdir, summary_t *sum)
{
	DIR		*dir;
	struct dirent	*file;

	if (!(dir = opendir(subdir)))
		qerr(1, "Cannot opendir `%s'", subdir);

	/* total = #filenames consisting of numbers */
	for (sum->total = 0; (file = readdir(dir)); )
	{
		bool		real = true;

		for (char *x = file->d_name; *x; x++)
			real &= *x >= '0' && *x <= '9';
		if (real)
		{
			struct stat	statbuf;
			char		*filename;

			(sum->total)++;
			asprintf(&filename, "%s/%s", subdir, file->d_name);
			if (!stat(filename, &statbuf))
				if (statbuf.st_mtime >= statbuf.st_atime)
					(sum->new)++;
			free(filename);
		}
	}

	closedir(dir);
	return true;
}

/**
 * Parse folder direcotry in MH or Maildir format
 *
 * \param[in]	dir	directory to parse
 * \param[out]	sum	folder summary
 *
 * \return		true if parsing succeeded
 */
bool
parse_dir(const char *dir, summary_t *sum)
{
	struct stat	filestat;
	char		*filename;
	bool		res;

	asprintf(&filename, "%s/cur", dir);
	if (lstat(filename, &filestat) < 0 || !S_ISDIR(filestat.st_mode))
		/* parse MH-style mailfolder */
		res = parse_mh(dir, sum);
	else
		/* parse cur/new/tmp-style mailfolder */
		res = parse_maildir(dir, sum);

	free(filename);
	return true;
}

/** Print number (of messages) or '-' if no messages */
static inline void
desc(const unsigned int i, const char *msg)
{
	if (!i)
		printf("%*c %s", spc, '-', msg);
	else
		printf("%*u %s", spc, i, msg);
}

/**
 * Print folder summary information
 *
 * \param[in]	folder	mailfolder name
 * \param[in]	sum	mailfolder summary
 */
void
display(const char *folder, const summary_t *sum)
{
	if (printfrm)
	{
		if (sum->new || (sum->total && printfrmold))
			puts(sum->frm);
		else
			printf("%s has no new mail\n", folder);
	}
	else if ((sum->total > 0) && (sum->new || !hideold))
	{
		printf("%-*s", mlen, folder);
		desc(sum->new, "new");
		if (oflag)
			desc(sum->flag, "flag");
		if (oanswer)
			desc(sum->answer, "answ");
		printf(" of total%*d\n", spc, sum->total);
	}
	else if (brief && sum->new > 0)
		printf("%s\n", folder);
}

/**
 * Parse any mailfolder
 * - Validate folder
 * - Determine folder format
 * - Call the correct parser routine
 * - Display results
 *
 * \param[in]	dir	directory containing mailfolders
 * \param[in]	filenm	mailfolder name
 *
 * \return		true if parsing succeeded
 */
bool
parse_folder(const char *dir, const char *filenm)
{
	struct stat	filestat;
	summary_t	sum = { 0, 0, 0, 0, NULL, 0 };
	char		*filename;

	asprintf(&filename, "%s/%s", dir, filenm);

	if (lstat(filename, &filestat) < 0)
		qerr(1, "Cannot stat `%s'", filename);

	/* Follow symlinks iff 'flink' */
	if (S_ISLNK(filestat.st_mode) && flink &&
			(stat(filename, &filestat) < 0))
		qerr(1, "Cannot stat `%s'", filename);

	if (printfrm)
	{
		sum.frmsz = 2048; /* arbitrary. may be extended later */
		sum.frm = malloc(sum.frmsz);
		sum.frm[0] = '\0';
	}

	if (S_ISDIR(filestat.st_mode))
	{
		bool parse_tree(char *maildir);

		/* parse directory mailfolder */
		if (parse_dir(filename, &sum))
			display(filenm, &sum);
		else
			/* recurse */
			parse_tree(filename);

		free(filename);
		free(sum.frm);
		return true;
	}
	if (!S_ISREG(filestat.st_mode) || !filestat.st_size)
	{
		free(filename);
		free(sum.frm);
		return false;
	}

	/* is it new? do we care? */
	if (hideunread)
	{
		if (brief || filestat.st_mtime < filestat.st_atime)
		{
			free(filename);
			free(sum.frm);
			return true;
		}
	}

	/* open and parse mailfolder */
	FILE		*file = fopen(filename, "r");
	const char	c = fgetc(file);
	if (!file)
		qerr(1, "Cannot open `%s'", filename);

	ungetc(c, file);
	parse_mbox_mmdf(file, &sum, '\1' == c);
	display(filenm, &sum);
	fclose(file);

	/* close and preserve timestamp */
	const struct utimbuf	filetime = {
		.actime = filestat.st_atime,
		.modtime = touch ? time(NULL) : filestat.st_mtime,
	};
	utime(filename, &filetime);

	free(filename);
	free(sum.frm);
	return true;
}

/**
 * Parse a directory with mailfolders
 * - Sort directory list
 * - Calls parse_folder() for each entry
 *
 * \param[in]	maildir	directory name
 *
 * \return		true if parsing succeeded
 */
bool
parse_tree(char *maildir)
{
	unsigned int	maxdir = 0,	/* max directory entries */
			maxfile = 0,	/* max filename length */
			count = 0;
	DIR		*dir;
	struct dirent	*file;

	/* sort files from dir */
	if (!(dir = opendir(maildir)))
		qerr(1, "Cannot opendir `%s'", maildir);

	while ((file = readdir(dir)))
	{
		const size_t	len = strlen(file->d_name) + 1;

		if (len > maxfile)
			maxfile = len;
		maxdir++;
	}

	/* dynamic size arrays ftw */
	char			dirlist[maxdir][maxfile];

	memset(dirlist, 0, maxdir * maxfile);

	rewinddir(dir);
	while ((file = readdir(dir)))
	{
		/* check if this mailfolder is worth reading */
		const size_t	len = strlen(file->d_name);
		char		fl = *file->d_name;

		if (showdot &&
			len &&
			'.' == *file->d_name &&
			file->d_name[1] != '\0' &&
			(file->d_name[2] != '\0' || file->d_name[1] != '.'))
		{
			/* .text is fine */
			fl = file->d_name[1];
		}
		if (len &&
			(fl < 'a' || fl > 'z') &&
			(fl < 'A' || fl > 'Z' || !caps) &&
			(fl < '0' || fl > '9'))
		{
			continue;
		}

		if ((len > 4) && !strcmp((file->d_name + len-5), ".lock"))
			continue;

		if (len > mlen)
			mlen = len;
		/* copy file to todo-list */
		strlcpy(dirlist[count++], file->d_name, maxfile);
	}
	closedir(dir);

	qsort(dirlist, maxdir, maxfile,
		(int (*)(const void *, const void *))strcmp);

	/* parse maildir */
	for (count = maxdir - count; count < maxdir; count++)
	{
		char	**folderp;

		/* ignore items on exclude list */
		for (folderp = exclude; *folderp; folderp++)
			if (!strcmp(*folderp, dirlist[count]))
				break;
		if (!*folderp)
			parse_folder(maildir, dirlist[count]);
	}

	return true;
}

/**
 * main loop
 * - Determine what the user wants
 * - Call parse_folder or parse_tree as appropriate
 */
int
main(int argc, char **argv)
{
	char		*maildir, *inbox_dir, *inbox_name, *remote;
	char		*slash;
	int		opt, optc;
	unsigned int	num_exclude;
	bool		inonly, inskip;
	char		**optv, *options;

 	maildir = remote = inbox_dir = inbox_name = NULL;
 	inonly = inskip = caps = false;
 	num_exclude = 0;

	/* nfrm copatibility */
	const size_t	len = strlen(argv[0]);

	if ((len > 5 && !strcmp(argv[0] + len - 5, "/nfrm")) ||
			(4 == len && !strcmp(argv[0], "nfrm")))
		inonly = printfrm = true;
	else if ((len > 4 && !strcmp(argv[0] + len - 4, "/frm")) ||
			(3 == len && !strcmp(argv[0], "frm")))
		inonly = printfrm = printfrmold = true;

	/* this is a dirty hack */
	if (getenv("MAIL_UNREAD_AS_OLD"))
		unreadasold = true;

	/* be nice to lazy people - insert defaults before cmdline */
	/* TODO: maybe put this in a configuration file some day? */
	if ((options = getenv("CHK4MAIL_OPTIONS")))
	{
		char	**ap;

		/* optv = { argv[0], options[0..], argv[1..] } */
		optv = (char **)malloc(argc + strlen(options));
		optv[0] = argv[0];
		ap = &optv[1];
		for (optc = 1; (*ap++ = strsep(&options, " \t")); optc++)
			/* DO NOTHING */;

		for (int i = 1; i < argc; i++)
			optv[optc++] = argv[i];
	}
	else
	{
		optc = argc;
		optv = argv;
	}

	exclude = (char **)malloc((size_t)optc);

	while ((opt = getopt(optc, optv, "#:acd:e:fhijlm:nqr:stux")) != EOF)
		switch (opt)
		{
		case '#':
			if ((spc = atoi(optarg)) <= 0)
				usage();
			break;
		case 'a':
			oanswer = true;
			break;
		case 'c':
			caps = true;
			break;
		case 'd':
			maildir = optarg;
			break;
		case 'e':
			exclude[num_exclude++] =
				'=' == optarg[0] || '+' == optarg[0]
				? optarg + 1
				: optarg;
			break;
		case 'f':
			oflag = true;
			break;
		case 'h':
			hideold = true;
			break;
		case 'i':
			inonly = true;
			break;
		case 'j':
			inskip = true;
			break;
		case 'l':
			flink = true;
			break;
		case 'm':
			if ((mlen = atoi(optarg)) <= 0)
				usage();
			break;
		case 'n':
			hideold = true;
			hideunread = true;
			break;
		case 'q':
			quiet = true;
			break;
		case 'r':
			remote = optarg;
			break;
		case 's':
			brief = true;
			break;
		case 't':
			touch = true;
			break;
		case 'u':
			unreadasold = true;
			break;
		case 'x':
			showdot = true;
			break;
		default:
			usage();
			/* NOTREACHED */
		}

	exclude[num_exclude] = NULL;

	if (brief)
		hideold = hideunread = true;

	/* run cmd on remote server */
	if (remote)
	{
		struct utsname	uts;

		if (uname(&uts) < 0)
			err(1, "uname failed");
		if ((slash = strchr(uts.nodename, '.')))
			*slash = '\0';
		if (strcmp(remote, uts.nodename))
		{
			while (--argc)
				*(argv[argc]-1) = ' ';
			/* execl() does not return on success */
			if (execl("/usr/bin/rsh", remote, *argv, NULL) < 0)
 				err(1, "Cannot exec rsh `%s'", remote);
		}
	}

	/* find inbox */
	if (!inskip)
	{
		char	*inbox = getenv("MAIL");

		if (!inbox)
		{
			inbox_dir = _PATH_MAILDIR;

			if (!(inbox_name = getenv("USER")))
			{
				const struct passwd	*pwent =
					getpwuid(getuid());

				if (pwent)
					inbox_name = pwent->pw_name;
				else
					/* failed to determine inbox */
					inskip = true;
			}
		}
		else if ((slash = strrchr(inbox, '/')))
		{
			if (!slash[1])
			{
				/* Maildir with trailing slash?
				 * Use dirname, not filename as inbox */
				*slash = '\0';
				slash = strrchr(inbox, '/');
				if (!slash)
					slash = inbox;
			}
			inbox_dir = strdup(inbox);
			inbox_dir[slash - inbox] = '\0';
			inbox_name = slash + 1;
		}
		else
		{
			inbox_dir = "";
			inbox_name = inbox;
		}
		if (inbox_name && strlen(inbox_name) > mlen)
			mlen = strlen(inbox_name);
	}

	/* Shortcut to the end */
	if (inonly)
		goto INBOX;

	/* find maildir */
	if (!maildir)
	{
		const char	*home = getenv("HOME");

		asprintf(&maildir, "%s/Mail", home);
		if (access(maildir, X_OK) < 0)
		{
			sprintf(maildir, "%s/mail", home); /* len ok */
			if (access(maildir, X_OK) < 0)
				/* fail later */
				sprintf(maildir, "%s/Mail", home); /* len ok */
		}
	}

	/* only parse specified mailboxes and exit */
	if (optind < argc)
	{
		/* always follow symlinks if filename is specified */
		flink = true;

		/* for a nice layout: find max filename length first */
		opt = optind;
		while (opt < argc)
		{
			const size_t nlen =
				(slash = strrchr(argv[opt], '/'))
				? strlen(slash + 1) 
				: strlen(argv[opt]);

			opt++;
			if (nlen > mlen)
				mlen = nlen;
		}
		opt = optind;
		/* all arguments are folder names */
		while (opt < argc)
		{
			if ('=' == *argv[opt] || '+' == *argv[opt])
				parse_folder(maildir, argv[opt] + 1);
			else if ((slash = strrchr(argv[opt], '/')))
			{
				*slash = '\0';
				parse_folder(argv[opt], slash + 1);
			}
			else
			{
				char	*cwd = getcwd(NULL, 0);

				if (cwd)
				{
					parse_folder(cwd, argv[opt]);
					free(cwd);
				}
			}
			opt++;
		}
		return 0;
	}

	/* sort and parse maildir */
	parse_tree(maildir);
	free(exclude);

INBOX:
	/* parse inbox */
	flink = true;
	if (!inskip && inbox_name)
		parse_folder(inbox_dir, inbox_name);

	return 0;
}
